#
# otojiro-fpw -- FreePWING script for OTOJIRO
#
# $Id: otojiro-fpw.pl 1.0.1.0 2002/09/11 04:39:01 satomii Exp $
# copyright (C) 2002, satomii. all rights reserved.
#
# this program is distributed in the hope that it will be useful, but
# without any warranty. see the GNU General Public License for the details.
#
# expected FreePWING version: 1.4.2 (or later)
#

use strict;
use Jcode;
use FreePWING::FPWUtils::FPWParser;

##-------------------------------------------------------------------------
## MAIN ROUTINE

# constants.
my $EUC_BULLET = "\xa1\xfe";		# 
my $EUC_CIRCLE = "\xa1\xfc";		# 
my $EUC_DIAMOND = "\xa2\xa1";		# 
my $EUC_RE_ANY = '[\xa1-\xfe][\xa0-\xfe]';

# global variables.
my $jcode = 'euc';
my ($fpwtext, $fpwheading, $fpwword2, $fpwcopyright, $wc);

# detect display character set.
if ($ENV{LANG} =~ /ja_JP\.([a-z]+)/i) {
	$jcode = lc($1);
	if ('eucjp' eq $jcode || 'ujis' eq $jcode) {
		$jcode = 'euc';
	} elsif ('sjis' ne $jcode
			 || 'jis' ne $jcode
			 || 'iso_2022_jp' ne $jcode) {
		$jcode = 'euc';
	}
} elsif ($ENV{OS} =~ /Windows/i
		 || $ENV{COMSPEC} =~ /cmd\.exe/i
		 || $ENV{COMSPEC} =~ /command\.com/i) {
	$jcode = 'sjis';
}

initialize_fpwparser('text' => \$fpwtext,
		'heading' => \$fpwheading,
		'word2' => \$fpwword2,
		'copyright' => \$fpwcopyright);

$wc = &make_body();
&write_copyright();

finalize_fpwparser('text' => \$fpwtext,
		'heading' => \$fpwheading,
		'word2' => \$fpwword2,
		'copyright' => \$fpwcopyright);

print "Total $wc entries were written.\n";
exit(0);

##-------------------------------------------------------------------------
## ENTRY CONSTRUCTION

sub make_body()
{
	my ($file, $word, $str, @lines);
	my $count = 0;

	foreach $file (@ARGV) {
		open(IN, "<$file") || die "Unable to open file: $file ($!)\n";
		print "Reading: $file\n";

		while (<IN>) {
			s/[ \t\r\n\x1a]+$//;
			next if ('' eq $_);

			$_ = Jcode::convert(sjis_han2zen($_), 'euc', 'sjis');
			if (!s/^\xa2\xa3//) {	# 
				die "Unexpected line in $file ($_)"
			}
			if (!s/^([^:]+) : //) {
				&error(0, "Unexpected line: $_");
				next;
			}

			$str = $1;
			$str =~ s/ *\([0-9]{3}[0-9]*\) *//g;

			if (!$word) {
				$word = $str;
			} elsif ($word ne $str) {
				&make_entry($word, @lines);
				$count++;
				$word = $str;
				@lines = ();
			}
			push(@lines, $_);
		}

		if (@lines) {
			&make_entry($word, @lines);
			$count++;
		}
		close(IN);
	}

	return $count;
}

sub make_entry($@)
{
	my ($word, @lines) = @_;
	my ($head_pos, $text_pos, $roman, $read);

	$fpwheading->new_entry()
		|| &error($fpwheading, "Unable to start new entry.");
	$fpwtext->new_entry()
		|| &error($fpwtext, "Unable to start new entry.");
	$head_pos = $fpwheading->entry_position();
	$text_pos = $fpwtext->entry_position();

	($roman, $read) = split(/ *; */, $word);
	print Jcode::convert("$read [$roman]\n", $jcode);

	$fpwheading->add_text($read)
		|| &error($fpwheading, "Unable to add heading: $read");
	$fpwword2->add_entry($read, $head_pos, $text_pos)
		|| &error($fpwword2, "Unable to add search word: $read");
	$fpwword2->add_entry($roman, $head_pos, $text_pos)
		|| &error($fpwword2, "Unable to add search word: $roman");

	if (!$fpwtext->add_keyword_start()
			|| !$fpwtext->add_text("$read [$roman]")
			|| !$fpwtext->add_keyword_end()
			|| !$fpwtext->add_newline()) {
		&error($fpwtext, "Unable to add text: $read [$roman]");
	}

	$fpwtext->add_indent_level(2)
		|| &error($fpwtext, "Unable to set indent (2).");

	if (0 < $#lines) {
		for (my $i = 0; $i <= $#lines; $i++) {
			&write_item($read, $i + 1, $lines[$i]);
		}
	} else {
		&write_item($read, 0, shift(@lines));
	}
}

sub write_item($$$)
{
	my ($read, $num, $item) = @_;
	my (@parts, $pos, $str);
	my $prefix = '';
	my $suffix = '';
	local $_;

	#if ($num) {
	#	$fpwtext->add_text("$num. ")
	#		|| &error($fpwtext, "Unable to write item number: $num");
	#}
	$fpwtext->add_text($EUC_BULLET)
		|| &error($fpwtext, "Unable to write bullet: $EUC_BULLET");

	# replace all "Q" to "`".
	$item = &euc_subst($item, "\xa1\xb2", "\xa1\xc1");
	$item =~ s/  +/ /g;

	# the first part should be the japanese description.
	$pos = &euc_index($item, $EUC_CIRCLE);
	if (-1 == $pos) {
		$_ = $item;
		$item = '';
	} else {
		$_ = substr($item, 0, $pos);
		$item = substr($item, $pos + 2);
	}

	# remove or move "..." at the end.
	while (1) {
		$pos = &euc_index($_, $EUC_DIAMOND);
		last if (-1 == $pos);

		$str = substr($_, $pos + 2);
		$_ = substr($_, 0, $pos);

		if ($str =~ /^\xa1\xd4(?:[\x20-\x7e]|$EUC_RE_ANY)+\xa1\xd5$/) {
			# s...t
			$prefix .= $str;
		} else {
			$suffix .= " $str";
		}
	}

	# move "G..." to the beginning of the line.
	$pos = &euc_index($_, "\xa1\xa8");
	if (-1 != $pos) {
		$str = &euc_subst(substr($_, $pos + 2), "\xa1\xa8", '');
		$_ = $str . substr($_, 0, $pos);
	}

	# conversational examples may lack the leading quatation mark.
	$pos = &euc_index($_, "\xa1\xd7");
	if (-1 != $pos) {
		my $pos0 = &euc_index($_, "\xa1\xd6");
		if (-1 == $pos0 || $pos < $pos0) {
			$_ = "\xa1\xd6" . $_;
		}
	}

	# write the description. "y$readz" is represented as bold text.
	$_ = &euc_subst($_, "\xa1\xda$read\xa1\xdb", "\n-!-$read\n");
	@parts = split(/\n/, $_);
	unshift(@parts, $prefix) if ($prefix);
	push(@parts, $suffix) if ($suffix);

	foreach $_ (@parts) {
		next if ('' eq $_);
		$_ = &euc_format($_);
		if (s/^-!-//) {
			$fpwtext->add_half_user_character('point')
				|| &error($fpwtext, "Unable to add user character: point");
		}
		$fpwtext->add_text($_) || &error($fpwtext, "Unable to add text: $_");
	}

	# the rest should contain one or more english expressions.
	if ($item) {
		$fpwtext->add_text('  ')
			|| &error($fpwtext, "Unable to write space '  '.");
		$item =~ s/ ~ / \xa1\xc1 /g;

		while ($item) {
			$pos = &euc_index($item, $EUC_CIRCLE);
			if (-1 == $pos) {
				$_ = $item;
				$item = '';
			} else {
				$_ = substr($item, 0, $pos);
				$item = substr($item, $pos + 2);
			}
			$_ = &euc_subst($_, $EUC_DIAMOND, " \xa2\xa8");
			$_ = &euc_format($_);
			$_ .= '; ' if ($item);
			$fpwtext->add_text($_)
				|| &error($fpwtext, "Unable to add text: $_");
		}
	}

	$fpwtext->add_newline() || &error($fpwtext, "Unable to add newline.");
}

sub write_copyright()
{
	my $user = defined($ENV{USER}) ? $ENV{USER} :
			(defined($ENV{USERNAME}) ? $ENV{USERNAME} : 'unknown user');

	my @time = localtime(time);
	@time = reverse(@time);
	@time = splice(@time, -6);
	$time[0] += 1900;
	$time[1] += 1;

	my @lines = (
		'Y FreePWING ',
		'',
		'̏Ђ́AElectronic Dictionary Project (EDP) 삵[E[ԌꎫuYvAFreePWING XNvg otojiro-fpw gp JIS X 4081 `ɕϊ̂łB̏Ђ̒쌠їṕAIWi̎̒쌠їpɏ̂Ƃ܂B',
		'',
		'Y̎f[^̈ꕔ܂͑SO҂ɃRs[AzzA]ڂ邱Ƃ͋ւĂ܂BÝAl͈͓̔ɂpwKEE|Ȃǂ̖ړIɗpł܂BcړIŗp邱Ƃ͂ł܂B',
		'',
		'̎̐ɎgꂽϊXNvg otojiro-fpw ̒쌠́A satomi ۗL܂B',
		'',
		sprintf("Generated by $user on %02d/%02d/%02d %02d:%02d:%02d", @time),
		'$Id: otojiro-fpw.pl 1.0.1.0 2002/09/11 04:39:01 satomii Exp $'
	);

	if (!$fpwcopyright->add_keyword_start()
			|| !$fpwcopyright->add_text(Jcode::convert($lines[0], 'euc'))
			|| !$fpwcopyright->add_keyword_end()
			|| !$fpwcopyright->add_newline()) {
		&error($fpwcopyright, "Unable to add coypright title: $lines[0]");
	}
	shift(@lines);

	while (@lines) {
		if ($lines[0]) {
			$fpwcopyright->add_text(Jcode::convert($lines[0], 'euc'))
				|| &error($fpwcopyright, "Unable to add text: $lines[0]");
		}
		$fpwcopyright->add_newline()
			|| &error($fpwcopyright, "Unable to add newline.");
		shift(@lines);
	}
}

##-------------------------------------------------------------------------
## JAPANESE STRING FUNCTIONS

sub sjis_han2zen($)
{
	my @chars = unpack('C*', $_[0]);
	my $ret = '';

	while (@chars) {
		my $c1 = shift(@chars);
		if (0xa1 == $c1) {								# B
			$ret .= "\x81\x42";
		} elsif (0xa2 == $c1 || 0xa3 == $c1) {			# uv
			$ret .= pack('CC', 0x81, $c1 - 0x2d);
		} elsif (0xa4 == $c1) {							# A
			$ret .= "\x81\x41";
		} elsif (0xa5 == $c1) {							# E
			$ret .= "\x81\x45";
		} elsif (0xa6 == $c1) {							# 
			$ret .= "\x83\x92";
		} elsif (0xa7 <= $c1 && $c1 <= 0xab) {			# @BDFH
			$ret .= pack('CC', 0x83, ($c1 - 0xa7) * 2 + 0x40);
		} elsif (0xac <= $c1 && $c1 <= 0xae) {			# 
			$ret .= pack('CC', 0x83, ($c1 - 0xac) * 2 + 0x83);
		} elsif (0xaf == $c1) {							# b
			$ret .= "\x83\x62";
		} elsif (0xb0 == $c1) {							# [
			$ret .= "\x81\x5b";
		} elsif (0xb1 <= $c1 && $c1 <= 0xb5) {			# ACEGI
			if (0xb3 == $c1 && 0xde == $chars[0]) {		# 
				$ret .= "\x83\x94";
				shift(@chars);
			} else {
				$ret .= pack('CC', 0x83, ($c1 - 0xb1) * 2 + 0x41);
			}
		} elsif (0xb6 <= $c1 && $c1 <= 0xc4) {			# J-g
			my $offset = 0;
			$offset++ if (0xc2 <= $c1);					# ceg
			if (0xde == $chars[0]) {					# K-h
				$offset++;
				shift(@chars);
			}
			$ret .= pack('CC', 0x83, ($c1 - 0xb6) * 2 + 0x4a + $offset);
		} elsif (0xc5 <= $c1 && $c1 <= 0xc9) {			# ijklm
			$ret .= pack('CC', 0x83, $c1 - 0xc5 + 0x69);
		} elsif (0xca <= $c1 && $c1 <= 0xce) {			# nqtwz
			my $offset = 0;
			if (0xde == $chars[0]) {					# orux{
				$offset = 1;
				shift(@chars);
			} elsif (0xdf == $chars[0]) {				# psvy|
				$offset = 2;
				shift(@chars);
			}
			$ret .= pack('CC', 0x83, ($c1 - 0xca) * 3 + 0x6e + $offset);
		} elsif (0xcf <= $c1 && $c1 <= 0xd3) {			# }~
			$ret .= pack('CC', 0x83,
					$c1 - 0xcf + 0x7d + ((0xd1 <= $c1) ? 1 : 0));
		} elsif (0xd4 <= $c1 && $c1 <= 0xd6) {			# 
			$ret .= pack('CC', 0x83, ($c1 - 0xd4) * 2 + 0x84);
		} elsif (0xd7 <= $c1 && $c1 <= 0xdb) {			# 
			$ret .= pack('CC', 0x83, $c1 - 0xd7 + 0x89);
		} elsif (0xdc == $c1) {							# 
			$ret .= "\x83\x8f";
		} elsif (0xdd == $c1) {							# 
			$ret .= "\x83\x93";
		} elsif (0xde == $c1 || 0xdf == $c1) {			# JK
			$ret .= pack('CC', 0x81, $c1 - 0x94);
		} elsif (0x81 <= $c1) {
			$ret .= pack('CC', $c1, shift(@chars));
		} else {
			$ret .= pack('C', $c1);
		}
	}
	return $ret;
}

sub euc_index($$)
{
	my ($str, $substr) = @_;
	my $len = length($substr);
	my $pos = 0;

	while ($str) {
		if ($substr eq substr($str, 0, $len)) {
			return $pos;
		} elsif ($str =~ s/^$EUC_RE_ANY//) {
			$pos += 2;
		} else {
			$pos++;
			$str = substr($str, 1);
		}
	}

	return -1;
}

sub euc_subst($$$)
{
	my ($str, $from, $to) = @_;
	my $len = length($from);
	my $ret = '';

	while ($str) {
		if ($from eq substr($str, 0, $len)) {
			$ret .= $to;
			$str = substr($str, $len);
		} elsif ($str =~ s/^($EUC_RE_ANY)//) {
			$ret .= $1;
		} else {
			$ret .= substr($str, 0, 1);
			$str = substr($str, 1);
		}
	}

	return $ret;
}

sub euc_format($)
{
	my @chars = unpack('C*', $_[0]);
	my $ret = '';

	while (@chars) {
		my $c1 = shift(@chars);
		if (0xa1 == $c1) {
			my $c2 = shift(@chars);
			if (0xca == $c2) {
				$ret .= '(';
			} elsif (0xcb == $c2) {
				$ret .= ')';
			} elsif (0xa9 == $c2) {
				$ret .= '?';
				if (@chars && $chars[0] != 0x20 && $chars[0] != 0xa1) {
					$ret .= ' ';
				}
			} elsif (0xaa == $c2) {
				$ret .= '!';
				if (@chars && $chars[0] != 0x20 && $chars[0] != 0xa1) {
					$ret .= ' ';
				}
			} elsif (0xa3 == $c2) {
				$ret .= pack('CC', $c1, $c2);
				if (0xa1 == $chars[0] && 0xa2 == $chars[1]) {
					splice(@chars, 0, 2);
				}
			} else {
				$ret .= pack('CC', $c1, $c2);
			}
		} elsif (0xa2 <= $c1 && $c1 <= 0xfe) {
			$ret .= pack('CC', $c1, shift(@chars));
		} else {
			$ret .= pack('C', $c1);
		}
	}

	return $ret;
}

##-------------------------------------------------------------------------
## ERROR GENERATION

sub error($$)
{
	my ($module, $file, $line) = caller;
	my ($fpw, $msg) = @_;
	my $err = $fpw ? $fpw->error_message() : '';
	my $period = ($msg =~ s/\.$//);

	$msg = Jcode::convert($msg, $jcode) if ($jcode);
	$msg = "Error: $file($line): $msg";
	$msg .= "($err)" if ($err);
	$msg .= $period if ($period);
	print STDERR "$msg\n";
	exit(2);
}

# Local Variables:
# encoding: japanese-shift-jis-dos
# tab-width: 4
# End:
